from dash import Dash, dcc, html, no_update, callback_context
from dash.dependencies import Input, Output, State


from . import ids
from ..utils.processor import process_indentation


def render(app: Dash) -> html.Div:
    @app.callback(
        [Output(ids.ADJUST_CHECKLIST, 'value'),
         Output(ids.FIT_CHECKLIST, 'value'),
         Output(ids.INDENTATION, 'value')],
        [Input(ids.ADJUST_CHECKLIST, 'value'),
         Input(ids.FIT_CHECKLIST, 'value')],
        State(ids.INDENTATION, 'value'),
        prevent_initial_call=True
    )
    def link_adjust_fit(adjust, fit, indentation):
        ctx = callback_context
        trigger_id = ctx.triggered[0]['prop_id'].split('.')[0]
        indentation = process_indentation(indentation)
        if trigger_id == ids.FIT_CHECKLIST and indentation == 0.0:
            return no_update, [], "Unable to proceed without indentation!"

        if trigger_id == ids.ADJUST_CHECKLIST and not fit:
            return no_update

        if trigger_id == ids.ADJUST_CHECKLIST and fit:
            fit = []

        if trigger_id == ids.FIT_CHECKLIST and not adjust:
            adjust = [True]

        return adjust, fit, no_update

    return html.Div(
        children=[
            html.Div(
                children=[
                    dcc.Checklist(
                        id=ids.FIT_CHECKLIST,
                        options=[
                            {'label': '  Show Fits', 'value': True}],
                        style={
                            'width': "25%",
                        },
                    ),
                    dcc.Input(
                        placeholder="Enter max indentation or interval...",
                        id=ids.INDENTATION,
                        style={
                            'width': "75%",
                        },
                    )
                ],
                style={
                    'display': 'flex',
                    'width': '100%',
                }
            ),
        ],
        style={
            'display': 'flex',
            'flex-direction': 'column',
            'width': '100%',
            'gap': '5px',
            'align-items': 'start',
        }
    )



from dash import Dash
from dash_bootstrap_components.themes import BOOTSTRAP

from .src.components.layout import create_layout


def run(debug: bool = False, port: int = 8050) -> None:
    app = Dash(__name__, external_stylesheets=[BOOTSTRAP])
    app.title = "AFM Data Analysis"
    app.layout = create_layout(app)

    app.run(debug=debug, port=str(port))



import numpy as np
from scipy.optimize import curve_fit
from sklearn.metrics import r2_score as r2


# Functions
def powerlaw(x, a, b):
    return a * (x ** b)


def poly(x, a, b, c):
    return a * (x ** b) + c * x


def hertzian(ind, diameter, e_star):
    return (4 / 3) * e_star * ((0.5 * diameter) ** 0.5) * (ind ** 1.5)


def exponential(t, y_0, tau, y_f):
    return (y_0 - y_f) * np.exp(-t * tau) + y_f


def biexponential(t, y_0, c, tau1, tau2, y_f):
    return (y_0 - y_f) * (c * np.exp(-t * tau1) + (1 - c) * np.exp(-t * tau2)) + y_f


def poroelastic(t, y_0, c, tau1, tau2, y_f):
    return (y_0 - y_f) * (c * np.exp(-np.sqrt(t) * tau1) + (1 - c) * np.exp(-t * tau2)) + y_f


# Fitting
def lin_fit(x, y):
    popt = np.polyfit(x, y, 1)
    y_pred = np.poly1d(popt)(x)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def calculate_residuals(y_true, y_pred):
    return np.sum((y_true - y_pred) ** 2)


def powerlaw_fit(x, y):
    x_min = min(x)
    y_min = min(y)
    x = (x - x_min)
    x_max = max(x)
    x = x / x_max
    y = (y - y_min)
    y_max = max(y)
    y = y / y_max
    bounds = ([0, 0], [np.inf, np.inf])
    popt, _ = curve_fit(powerlaw, x, y, bounds=bounds)
    y_pred = powerlaw(x, *popt)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred * y_max + y_min


def poly_fit(x, y):
    x_min = min(x)
    y_min = min(y)
    x = (x - x_min)
    x_max = max(x)
    x = x / x_max
    y = (y - y_min)
    y_max = max(y)
    y = y / y_max
    bounds = ([0, 0, 0], [np.inf, np.inf, np.inf])
    p0 = [1, 1, 1]
    popt, _ = curve_fit(poly, x, y, bounds=bounds, p0=p0, jac='3-point')
    y_pred = poly(x, *popt)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred * y_max + y_min


def hertzian_fit(x, y, diameter):
    def wrapper(ind, e_star):
        return hertzian(ind, diameter, e_star)

    popt, _ = curve_fit(wrapper, x, y)
    y_pred = hertzian(x, diameter, *popt)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def exponential_fit(x, y, bound=False):
    x -= x[0]
    if bound:
        def bnd_wrapper(t, tau):
            return exponential(t, y[0], tau, y[-1])

        bounds = ([0], [np.inf])
        popt, _ = curve_fit(bnd_wrapper, x, y, bounds=bounds)
        args = [y[0], *popt, y[-1]]
    else:
        def ubnd_wrapper(t, tau, y_f):
            return exponential(t, y[0], tau, y_f)

        bounds = ([0, -np.inf], [np.inf, np.inf])
        popt, _ = curve_fit(ubnd_wrapper, x, y, bounds=bounds)
        args = [y[0], *popt]

    y_pred = exponential(x, *args)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def biexponential_fit(x, y, bound=False):
    x -= x[0]
    if bound:
        def bnd_wrapper(t, c, tau1, tau2):
            return biexponential(t, y[0], c, tau1, tau2, y[-1])

        bounds = ([0, 0, 0], [1, np.inf, np.inf])
        p0 = [0.4, 1, 0.1]
        popt, _ = curve_fit(bnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        # args = [y[0], *popt, y[-1]]

        # y_pred = biexponential(x, *args)
        # r2_score = r2(y, y_pred)

    else:
        def ubnd_wrapper(t, c, tau1, tau2, y_f):
            return biexponential(t, y[0], c, tau1, tau2, y_f)

        results = {}

        p0 = [0.4, 1, 0.1, y[-1]]
        bounds = ([0, 0, 0, -np.inf], [1, np.inf, np.inf, np.inf])

        popt_with_p0, _ = curve_fit(
            ubnd_wrapper, x, y, bounds=bounds, p0=p0, method='trf', jac='3-point')
        y_pred_with_p0 = biexponential(x, y[0], *popt_with_p0)
        r2_with_p0 = r2(y, y_pred_with_p0)
        results['with_p0'] = (popt_with_p0, r2_with_p0, y_pred_with_p0)

        popt_without_p0, _ = curve_fit(
            ubnd_wrapper, x, y, bounds=bounds, method='trf', jac='3-point')
        y_pred_without_p0 = biexponential(x, y[0], *popt_without_p0)
        r2_without_p0 = r2(y, y_pred_without_p0)
        results['without_p0'] = (
            popt_without_p0, r2_without_p0, y_pred_without_p0)

        if results['with_p0'][1] >= results['without_p0'][1]:
            best_fit = results['with_p0']
        else:
            best_fit = results['without_p0']

        popt, r2_score, y_pred = best_fit
        '''
        # y0 = y[0]
        y_f = y[-1]
        # e_threshold = y0 - 0.63 * (y0 - y_f)
        # e_time = x[np.where(y < e_threshold)[0][0]]
        # tau1_guess = e_time
        # tau2_guess = 0.1 * tau1_guess

        p0 = [0.4, 1, 0.1, y_f]
        bounds = ([0, 0, 0, -np.inf], [1, np.inf, np.inf, np.inf])
        
        popt, _ = curve_fit(ubnd_wrapper, x, y,
                            bounds=bounds,
                            p0=p0,
                            method='trf',
                            jac='3-point'
                            )
        args = [y[0], *popt]
        '''

    # y_pred = biexponential(x, *args)
    # r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def poroelastic_fit(x, y, bound=False):
    x -= x[0]
    if bound:
        def bnd_wrapper(t, c, tau1, tau2):
            return poroelastic(t, y[0], c, tau1, tau2, y[-1])

        bounds = ([0, 0, 0], [1, np.inf, np.inf])
        p0 = [0.4, 1, 0.1]
        popt, _ = curve_fit(bnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        args = [y[0], *popt, y[-1]]
    else:
        def ubnd_wrapper(t, c, tau1, tau2, y_f):
            return poroelastic(t, y[0], c, tau1, tau2, y_f)

        p0 = [0.4, 1, 0.1, 0]
        bounds = ([0, 0, 0, -np.inf], [1, np.inf, np.inf, np.inf])
        popt, _ = curve_fit(ubnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        args = [y[0], *popt]

    y_pred = poroelastic(x, *args)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


if __name__ == '__main__':
    pass



import json
import zipfile
import pickle
import io

from dash import Dash, dcc, html, no_update
from dash.dependencies import Input, Output, State

from ..components import ids
from ..utils.utils import load, save_afm, load_afm
from ..utils.processor import (
    process_experiment,
    process_images,
    process_indentation,
    process_pixel,
    get_pdf,
    merge_images
)


def render(app: Dash) -> html.Div:
    @app.callback(
        Output(ids.DOWNLOAD, 'data', allow_duplicate=True),
        Input(ids.DOWNLOAD_ANNOTATIONS, 'n_clicks'),
        [State(ids.CP_ANNOTATIONS, 'data'),
         State(ids.VD_ANNOTATIONS, 'data'),
         State(ids.IM_ANNOTATIONS, 'data'),
         State(ids.EXPERIMENT_PATH, 'value')],
        prevent_initial_call=True
    )
    def download_annotations(_, cp_data, vd_data, im_data, experiment_path):
        exp_name = experiment_path.split('\\')[-1]

        annotations = {
            'cp_annotations.json': cp_data,
            'vd_annotations.json': vd_data,
            'im_annotations.json': im_data
        }

        buffer = io.BytesIO()
        with zipfile.ZipFile(buffer, 'w') as zip_file:
            for filename, data in annotations.items():
                data_dict = json.loads(data)
                json_string = json.dumps(data_dict, indent=4)
                zip_file.writestr(f'{exp_name}_{filename}', json_string)
        buffer.seek(0)

        return dcc.send_bytes(buffer.getvalue(), f'{exp_name}_annotations.zip')

    @app.callback(
        [Output(ids.DOWNLOAD, 'data', allow_duplicate=True),
         Output(ids.EXPERIMENT_CACHE, 'data', allow_duplicate=True),
         Output(ids.DOWNLOAD_FITS, 'children', allow_duplicate=True),
         Output(ids.INDENTATION, 'value', allow_duplicate=True)],
        Input(ids.DOWNLOAD_FITS, "n_clicks"),
        [State(ids.LOAD_EXPERIMENT, 'n_clicks'),
         State(ids.EXPERIMENT_CACHE, 'data'),
         State(ids.IMAGES, 'data'),
         State(ids.CP_ANNOTATIONS, 'data'),
         State(ids.VD_ANNOTATIONS, 'data'),
         State(ids.IM_ANNOTATIONS, 'data'),
         State(ids.INDENTATION, 'value'),
         State(ids.PIXEL_SIZE, 'value'),
         State(ids.EXPERIMENT_PATH, 'value')],
        prevent_initial_call=True
    )
    def download_fits_csv(_, n_clicks, experiment_cache, encoded_images, cp_data, vd_data, im_data, indentation, pixel, experiment_path):

        if indentation:
            experiment = load_afm(experiment_cache['raw'])
            indentation = process_indentation(indentation)
            experiment_processed, df_fits = process_experiment(
                experiment, cp_data, vd_data, indentation)
            exp_name = experiment_path.split('\\')[-1]

            images: dict = load(encoded_images)
            keys = experiment_processed.experiment.keys()
            pixel = process_pixel(pixel)
            df_images = process_images(images, keys, im_data, pixel)

            df = df_fits.join(df_images)

            processed_file_path = save_afm(
                experiment_cache['processed'], experiment_processed, name='processed')  # cant do this if you have processed it already
            experiment_cache['processed'] = processed_file_path

            return dcc.send_data_frame(df.to_csv, filename=f"{exp_name}_fits.csv"), experiment_cache, no_update, no_update

        return no_update, no_update, no_update, "Unable to proceed without indentation!"

    @app.callback(
        [Output(ids.DOWNLOAD, 'data', allow_duplicate=True),
         Output(ids.DOWNLOAD_CURVES, 'children')],
        Input(ids.DOWNLOAD_CURVES, "n_clicks"),
        [State(ids.EXPERIMENT_CACHE, 'data'),
         State(ids.EXPERIMENT_PATH, 'value')],
        prevent_initial_call=True
    )
    def download_curves_pdf(_, experiment_cache, experiment_path):
        experiment_processed = load_afm(experiment_cache['processed'])
        exp_name = experiment_path.split('\\')[-1]

        pdf_merger = experiment_processed.experiment.export_figures()
        pdf_src = get_pdf(pdf_merger)

        return dcc.send_bytes(src=pdf_src.getvalue(), filename=f"{exp_name}_curves.pdf", base64=True), no_update

    @app.callback(
        [Output(ids.DOWNLOAD, 'data', allow_duplicate=True),
         Output(ids.DOWNLOAD_IMAGES, 'children')],
        Input(ids.DOWNLOAD_IMAGES, "n_clicks"),
        [State(ids.IMAGES, 'data'),
         State(ids.IM_ANNOTATIONS, 'data'),
         State(ids.EXPERIMENT_PATH, 'value')],
        prevent_initial_call=True
    )
    def download_images_pdf(_, encoded_images, im_data, experiment_path):
        exp_name = experiment_path.split('\\')[-1]

        images: dict = load(encoded_images)

        pdf_merger = merge_images(images, im_data)
        pdf_src = get_pdf(pdf_merger)

        return dcc.send_bytes(src=pdf_src.getvalue(), filename=f"{exp_name}_images.pdf", base64=True), no_update

    @app.callback(
        [Output(ids.DOWNLOAD, 'data', allow_duplicate=True),
         Output(ids.DOWNLOAD_EXPERIMENT, 'children')],
        Input(ids.DOWNLOAD_EXPERIMENT, "n_clicks"),
        [State(ids.EXPERIMENT_CACHE, 'data'),
         State(ids.EXPERIMENT_PATH, 'value')],
        prevent_initial_call=True
    )
    def download_experiment(_, experiment_cache, experiment_path):
        exp_name = experiment_path.split('\\')[-1]

        experiment = load_afm(experiment_cache['raw'])
        try:
            experiment_processed = load_afm(experiment_cache['processed'])
            if experiment_processed:
                experiment = experiment_processed
        except:
            pass

        return dcc.send_bytes(src=pickle.dumps(experiment), filename=f"{exp_name}.afm", base64=True), no_update

    return html.Div(
        children=[
            dcc.Download(id=ids.DOWNLOAD),
            dcc.Loading(
                id=ids.DOWNLOAD_ANIMATION,
                type="dot",
                children=html.Div(
                    children=[
                        html.Button(
                            children="Download Fits",
                            id=ids.DOWNLOAD_FITS,
                            n_clicks=0,
                            className="dash-button"
                        ),
                        html.Button(
                            children="Download Curves",
                            id=ids.DOWNLOAD_CURVES,
                            n_clicks=0,
                            className="dash-button"
                        ),
                        html.Button(
                            children="Download Images",
                            id=ids.DOWNLOAD_IMAGES,
                            n_clicks=0,
                            className="dash-button"
                        ),
                    ],
                    style={
                        'display': 'flex',
                        'gap': '5px',
                        'align-items': 'start',
                    }
                )
            ),
            html.Button(
                children="Download Experiment",
                id=ids.DOWNLOAD_EXPERIMENT,
                n_clicks=0,
                className="dash-button"
            ),
        ],
        style={
            'display': 'flex',
            'flex-direction': 'column',
            'width': '100%',
            'gap': '5px',
            'align-items': 'start',
        }
    )



import json

from dash import Dash, html, no_update
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State

from . import ids, image
from ..utils.utils import load, get_current_annotation, parse_path


def render(app: Dash) -> html.Div:
    @app.callback(
        Output(ids.IMAGE_HOLDER, 'figure', allow_duplicate=True),
        [Input(ids.CURVE_DROPDOWN, 'value'),
         Input(ids.IM_ANNOTATIONS, 'data')],
        State(ids.IMAGES, 'data'),
        prevent_initial_call=True
    )
    def update_image_frame(curve_value, im_data, encoded_images):
        if not curve_value:
            raise PreventUpdate

        key = eval(curve_value)['key']
        new_key = key[:-1] if len(key) > 1 else key

        images: dict = load(encoded_images)

        if new_key in images:
            im = get_current_annotation(curve_value, im_data)
            # THIS ONLY SHOWS THE FIRST IMAGE IF THERE IS ONE
            image_path = images[new_key][0]
            img = image.handle_image(image_path, im)
            return img
        else:
            return image.make()

    @app.callback(
        [Output(ids.IM_ANNOTATIONS, 'data', allow_duplicate=True),
         Output(ids.IMAGE_HOLDER, 'clickData')],
        Input(ids.IMAGE_HOLDER, 'clickData'),
        [State(ids.CURVE_DROPDOWN, 'value'),
         State(ids.IM_ANNOTATIONS, 'data'),
         State(ids.IMAGE_HOLDER, 'figure')],
        prevent_initial_call=True
    )
    def handle_click(clickData, curve_value, im_data, fig):
        key = eval(curve_value)['key']
        im_annotations = json.loads(im_data)
        selection = im_annotations[repr(key)]['selection']
        if selection == 'auto':
            x, y = clickData['points'][0]['y'], clickData['points'][0]['x']
            im_annotations[repr(key)]['clickData'] = [[x, y]]
            return json.dumps(im_annotations), None

        if selection == 'manual' and 'shapes' in fig['layout']:
            shapes = fig['layout']['shapes'][0]
            im_annotations[repr(key)]['clickData'] = parse_path(shapes['path'])
            fig['layout']['shapes'] = []
            return json.dumps(im_annotations), None

        return no_update, None

    return html.Div(
        className='image',
        children=[
            image.render(ids.IMAGE_HOLDER)
        ],
        style={
            'height': '280px',
            'width': 'auto',
        },
    )



import ast

import json
import io

import pandas as pd
import PyPDF2 as pdf

from tqdm import tqdm
from io import BytesIO
from PyPDF2 import PdfMerger

from ..components.image import read, get_cell_info, process_image, handle_image
from ...afm import AFM


def process_experiment(experiment: AFM, cp_data: str, vd_data: str, indentation: float | list[float]) -> tuple[AFM, pd.DataFrame]:
    cp_annotations = json.loads(cp_data)
    cp_annotations = {ast.literal_eval(
        key): value for key, value in cp_annotations.items()}

    vd_annotations = json.loads(vd_data)
    vd_annotations = {ast.literal_eval(
        key): value for key, value in vd_annotations.items()}

    experiment.experiment.update_annotations(cp_annotations)
    experiment.experiment.update_annotations(vd_annotations)

    df = experiment.experiment.get_fit_all(
        experiment.probe_diameter, ind=indentation)
    return experiment, df


def process_indentation(indentation: str) -> float | list[float]:
    try:
        indentation_list = indentation.split(';')
        ind = [float(i) for i in indentation_list]
    except (ValueError, AttributeError):
        return 0.0

    if len(ind) == 1:
        ind = ind[0]
    return ind


def process_pixel(pixel: str) -> float:
    try:
        pixel_size = float(pixel)
    except (ValueError, TypeError):
        return 1.0

    return pixel_size


def get_pdf(pdf_merger: PdfMerger) -> BytesIO:
    pdf_bytes = BytesIO()
    pdf_merger.write(pdf_bytes)
    pdf_bytes.seek(0)

    return pdf_bytes


def process_images(images: dict, keys: list[tuple[str, ...]], im_data: str, pixel: float = 1.0) -> pd.DataFrame:
    cell_info_list = []

    for key in keys:
        cell_info = get_cell_info([])
        new_key = key[:-1] if len(key) > 1 else key

        if new_key in images:
            annotations = json.loads(im_data)
            im = annotations[repr(key)]

            image_path = images[new_key][0]
            image_array = read(image_path)
            cell_contours = process_image(image_array, im)

            if len(cell_contours) == 1:
                cell_info = get_cell_info(cell_contours[0], pixel)

        cell_info_list.append(cell_info)

    return pd.DataFrame(cell_info_list)


def merge_images(images: dict, im_data: str):
    merger = pdf.PdfMerger()
    annotations = json.loads(im_data)

    for key, im in tqdm(annotations.items()):
        key = eval(key)
        new_key = key[:-1] if len(key) > 1 else key
        for image_key, image_paths in images.items():
            if image_key != new_key:
                continue

            # THIS ONLY SHOWS THE FIRST IMAGE IF THERE IS ONE
            image_path = image_paths[0]
            img = handle_image(image_path, im)

            title = ("").join(image_path.split('\\')[-1].split('.')[:-1])
            title = fr"$\text{{{title}}}$"
            img.update_layout(
                title={
                    'text': title,
                    'x': 0.5,
                    'y': 0.95,
                    'xanchor': 'center'
                }
            )
            img_pdf = io.BytesIO(img.to_image(format='pdf'))
            merger.append(img_pdf)

    return merger



import numpy as np
import numpy.typing as npt
from typing import Callable

import cv2
from skimage import filters
from skimage.morphology import closing, square, remove_small_objects
from skimage.feature import canny
from skimage.measure import label as _label
from scipy import ndimage as ndi

import plotly.graph_objects as go
from dash import dcc


def render(id: str) -> dcc.Graph:
    return dcc.Graph(
        id=id,
        figure=make(),
        style={
            'display': 'flex',
            'height': '100%',
        }
    )


def plot(img: go.Figure, image_array: npt.NDArray[np.uint8]) -> go.Figure:
    img.add_trace(go.Image(z=image_array, hovertemplate="(%{x}, %{y})<extra></extra>"))
    return img


def make() -> go.Figure:
    img = go.Figure()
    img.update_layout(
        plot_bgcolor='white',
        paper_bgcolor='white',
        newshape_line_color='blue',
        newshape_line_width=3,
        newshape_opacity=0.5,
        activeshape=dict(fillcolor='rgba(0, 255, 0, 0.2)'),
        margin=dict(t=10, b=10, l=10, r=10),
        xaxis=dict(visible=False),
        yaxis=dict(visible=False),
    )
    return img


def read(image_path: str) -> npt.NDArray[np.uint8]:
    image = cv2.imread(image_path, cv2.IMREAD_UNCHANGED)
    return image.astype(np.uint8)


def _contrast(gray: npt.NDArray[np.uint8], clipLimit=2.0) -> npt.NDArray[np.uint8]:
    rgb = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)
    lab = cv2.cvtColor(rgb, cv2.COLOR_BGR2LAB)
    l_channel, a, b = cv2.split(lab)

    clahe = cv2.createCLAHE(clipLimit=clipLimit, tileGridSize=(8, 8))
    cl = clahe.apply(l_channel)

    limg = cv2.merge((cl, a, b))
    enhanced_img = cv2.cvtColor(limg, cv2.COLOR_LAB2BGR)
    return cv2.cvtColor(enhanced_img, cv2.COLOR_BGR2GRAY).astype(np.uint8)


def _filter(gray: npt.NDArray[np.uint8], func: Callable) -> npt.NDArray[np.uint8]:
    thr = func(gray)
    thr = 127 - abs(thr - 127)
    mask = (gray < thr) * 255
    mask = mask.astype(np.uint8)
    return mask


def label(image: npt.NDArray[np.uint8], clipLimit=2.0, sq: int = 5) -> npt.NDArray[np.uint8]:
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY).astype(np.uint8)
    gray = _contrast(gray, clipLimit)
    closed = closing(gray, square(sq))
    fg_gray = (closed - gray)
    filtered = _filter(fg_gray, filters.threshold_triangle)
    edges = canny(filtered)
    filled = ndi.binary_fill_holes(edges)
    cleaned = remove_small_objects(filled, 10)
    labels = np.array(_label(cleaned), dtype=np.uint8)
    return labels


def _contours(mask: npt.NDArray[np.uint8], count: int = 100) -> list[list[list[int]]]:
    cell_contours, _ = cv2.findContours(
        mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    sorted_contours = sorted(
        cell_contours, key=cv2.contourArea, reverse=True)[:count]
    areas = np.array([cv2.contourArea(contour)
                      for contour in sorted_contours])
    valid_indices = np.where(0.2 < (areas / np.max(areas)))[0]
    cell_contours = [sorted_contours[i] for i in valid_indices]
    return cell_contours


def _clean(image: npt.NDArray[np.uint8], cell_contour: list[list[int]]) -> list[list[list[int]]]:
    cell_mask = np.zeros_like(image)
    cell_hull = cv2.convexHull(np.array(cell_contour))
    cv2.fillPoly(cell_mask, [cell_hull], (255, 255, 255))

    # , shrink: int = 0
    # if shrink > 0:
    #     kernel_size = 5
    #     kernel = np.ones((kernel_size, kernel_size), np.uint8)
    #     cell_mask = cv2.erode(
    #         cell_mask, kernel, iterations=shrink).astype(np.uint8)

    mask = cell_mask.astype(np.uint8)
    cell_contours = _contours(mask)
    return cell_contours


def _mask(image: npt.NDArray[np.uint8], mask: npt.NDArray[np.uint8], alpha: float = 0.5) -> npt.NDArray[np.uint8]:
    masked_pixels = np.where(mask == 255)
    image[masked_pixels] = (alpha * mask[masked_pixels] +
                            (1 - alpha) * image[masked_pixels])
    return image


def get_contours(labels: npt.NDArray[np.uint8], clean: bool = True) -> list[list[list[int]]]:
    cell_contours = _contours(labels)
    if len(cell_contours) == 1 and clean:
        cell_contours = _clean(labels, cell_contours[0])
    return cell_contours


def add_contours(image: npt.NDArray[np.uint8], cell_contours: list[list[list[int]]]) -> npt.NDArray[np.uint8]:
    if len(cell_contours) == 1:
        cell_mask = np.zeros_like(image)
        cv2.fillPoly(cell_mask, [np.array(cell_contours[0])], (255, 0, 0))
        image = _mask(image, cell_mask)

    for contour in cell_contours:
        cv2.drawContours(image, [np.array(contour)], -1, (255, 0, 0), 1)
    return image


def handle_image(image_path: str, im: dict) -> go.Figure:
    img = make()
    image_array = read(image_path)
    cell_contours = process_image(image_array, im)
    image_array = add_contours(image_array, cell_contours)
    img = plot(img, image_array)
    if im['selection'] == 'manual':
        img.update_layout(modebar_add=['drawclosedpath', 'eraseshape'])
    return img


def process_image(image_array: npt.NDArray[np.uint8], im: dict) -> list[list[list[int]]]:
    selection = im['selection']
    clickData = im['clickData']
    cell_contours = clickData

    if selection == 'auto':
        clipLimit = im['contrast']
        sq = im['size']
        labels = label(image_array, clipLimit=clipLimit, sq=sq)

        if clickData:
            x, y = clickData[0]
            labels[labels != labels[x, y]] = 0
            labels[labels == labels[x, y]] = 1

        cell_contours = get_contours(labels)

    return cell_contours


def get_ellipse(contour: list[list[int]]) -> cv2.typing.RotatedRect:
    ellipse = cv2.fitEllipse(np.array(contour))
    return ellipse


def cell_info() -> dict:
    info = dict()
    info['eq_diameter'] = 0
    info['ave_diameter'] = 0
    info['perimeter'] = 0
    info['area'] = 0
    info['circularity'] = 0
    info['aspect_ratio'] = 0
    info['taylor_ratio'] = 0
    info['eccentricity'] = 0
    return info


def get_cell_info(contour: list[list[int]], pixel: float = 1.0) -> dict:
    info = cell_info()

    if len(contour) > 1:
        perimeter = cv2.arcLength(np.array(contour), closed=True)
        area = cv2.contourArea(np.array(contour))
        eq_diameter = 2 * np.sqrt(area / np.pi)
        circularity = 4 * np.pi * area / (perimeter ** 2)

        ellipse = get_ellipse(contour)
        _, (d1, d2), _ = ellipse

        d1, d2 = max(d1, d2), min(d1, d2)
        ave_diameter = (d1 + d2) / 2
        aspect_ratio = d1 / d2
        taylor_ratio = (d1 - d2) / (d1 + d2)
        eccentricity = np.sqrt(1 - d2 / d1)

        info['eq_diameter'] = eq_diameter * pixel
        info['ave_diameter'] = ave_diameter * pixel
        info['perimeter'] = perimeter * pixel
        info['area'] = area * (pixel ** 2)
        info['circularity'] = circularity
        info['aspect_ratio'] = aspect_ratio
        info['taylor_ratio'] = taylor_ratio
        info['eccentricity'] = eccentricity

    return info


def correct(angle: float) -> float:
    if angle > 90:
        angle = angle - 90
    else:
        angle = angle + 90

    return angle


def draw_radius(
    image: npt.NDArray[np.uint8],
    center: tuple[float, float],
    angle: float,
    radius: float,
    color: tuple[int, int, int]
) -> npt.NDArray[np.uint8]:
    xc, yc = center
    x1 = xc + np.cos(np.radians(angle)) * radius
    y1 = yc + np.sin(np.radians(angle)) * radius
    x2 = xc + np.cos(np.radians(angle + 180)) * radius
    y2 = yc + np.sin(np.radians(angle + 180)) * radius
    cv2.line(image, (int(x1), int(y1)), (int(x2), int(y2)), color, 1)
    return image


def draw_ellipse(
    image: npt.NDArray[np.uint8],
    ellipse: cv2.typing.RotatedRect
) -> npt.NDArray[np.uint8]:
    (xc, yc), (d1, d2), angle = ellipse
    cv2.ellipse(image, ellipse, (0, 255, 0), 1)

    rmajor = max(d1, d2) / 2
    angle = correct(angle)
    image = draw_radius(image, (xc, yc), angle, rmajor, (0, 0, 255))
    rminor = min(d1, d2) / 2
    angle = correct(angle)
    image = draw_radius(image, (xc, yc), angle, rminor, (255, 0, 0))
    return image



from dash import Dash, dcc, html, no_update, callback_context
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
import json

from . import ids


def render(app: Dash) -> html.Div:
    @app.callback(
        [Output(ids.IM_ANNOTATIONS, 'data', allow_duplicate=True),
         Output(ids.CLIP_LIMIT, 'value'),
         Output(ids.SQUARE_VALUE, 'value'),
         Output(ids.IM_DROPDOWN, 'value')],
        [Input(ids.CURVE_DROPDOWN, 'value'),
         Input(ids.CLIP_LIMIT, 'value'),
         Input(ids.SQUARE_VALUE, 'value'),
         Input(ids.IM_DROPDOWN, 'value'),
         Input(ids.IM_ANNOTATIONS, 'data')],
        prevent_initial_call=True
    )
    def update_im(curve_value, clipLimit, sq, im_value, im_data):
        if not curve_value:
            raise PreventUpdate

        ctx = callback_context
        trigger_id = ctx.triggered[0]['prop_id'].split('.')[0]

        key = eval(curve_value)['key']
        im_annotations = json.loads(im_data)
        im = im_annotations[repr(key)]
        if trigger_id in (ids.CURVE_DROPDOWN, ids.IM_ANNOTATIONS):
            selection = im['selection']
            if selection == 'auto':
                clipLimit = im['contrast']
                sq = im['size']
                return no_update, clipLimit, sq, selection
            else:
                return no_update, no_update, no_update, selection

        im['selection'] = im_value
        im['clickData'] = []
        if im_value == 'auto':
            im['contrast'] = clipLimit
            im['size'] = sq
        if im_value == 'manual' and im.get('contrast') and im.get('size'):
            del im['contrast']
            del im['size']
        im_annotations[repr(key)] = im
        return json.dumps(im_annotations), no_update, no_update, im_value

    @app.callback(
        [Output(ids.CLIP_LIMIT, 'disabled'),
         Output(ids.SQUARE_VALUE, 'disabled')],
        Input(ids.IM_DROPDOWN, 'value')
    )
    def update_input_state(selected_value):
        if selected_value == 'auto':
            return False, False
        else:
            return True, True

    return html.Div(
        children=[
            html.Div(children="Contrast: "),
            dcc.Input(
                id=ids.CLIP_LIMIT,
                type='number',
                value=1.5,
                step=0.1,
                min=0.1,
                style={'width': '25%'}
            ),
            html.Div(children="Size: "),
            dcc.Input(
                id=ids.SQUARE_VALUE,
                type='number',
                value=5,
                step=1,
                min=2,
                style={'width': '25%'}
            ),
            dcc.Input(
                id=ids.PIXEL_SIZE,
                type='text',
                placeholder='Pixel size...',
                style={'width': '50%'}
            ),
            dcc.Dropdown(
                id=ids.IM_DROPDOWN,
                options=[
                    {'label': 'Exclude', 'value': 'exclude'},
                    {'label': 'Auto', 'value': 'auto'},
                    {'label': 'Manual', 'value': 'manual'}
                ],
                value='exclude',
                clearable=False,
                style={'width': '99%'}
            ),
        ],
        style={
            'display': 'flex',
            'width': '100%',
            'gap': '5px',
            'align-items': 'start',
        }
    )
